import time

from lakefs.branch import Branch
from lakefs.repository import Repository

from lakefs_spec import LakeFSFileSystem
from lakefs_spec.spec import md5_checksum
from tests.util import RandomFileFactory, with_counter


def test_checksum_matching(
    random_file_factory: RandomFileFactory,
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    random_file = random_file_factory.make()

    fs.client, counter = with_counter(fs.client)

    lpath = str(random_file)
    rpath = f"{repository.id}/{temp_branch.id}/{random_file.name}"
    fs.put_file(lpath, rpath)

    # assert that MD5 hash is insensitive to the block size
    blocksizes = [2**5, 2**8, 2**10, 2**12, 2**22]
    for blocksize in blocksizes:
        local_checksum = md5_checksum(lpath, blocksize)
        assert local_checksum == fs.checksum(rpath)
        # this test sometimes fails because of a race condition in the client
        time.sleep(0.1)

    # we expect to get one `info` call per upload attempt, but only one actual upload.
    assert counter.count("objects_api.stat_object") == len(blocksizes) + 1


def test_checksum_directory(
    fs: LakeFSFileSystem,
    repository: Repository,
) -> None:
    rpath = f"lakefs://{repository.id}/main/data/"
    assert fs.isdir(rpath)
    assert fs.checksum(rpath) is None, "Checksum of a directory should be None"



from lakefs import Branch, Repository

from lakefs_spec.spec import LakeFSFileSystem


def test_gh_299(
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    """
    Regression test for GitHub issue 299: Extending ref expression validity.
    https://github.com/aai-institute/lakefs-spec/issues/299
    """

    prefix = f"lakefs://{repository.id}/{temp_branch.id}"
    datapath = f"{prefix}/data.txt"

    # add new file, and immediately commit.
    fs.pipe(datapath, b"data1")
    temp_branch.commit(message="Add data.txt")

    # update the file with new data, commit again.
    fs.pipe(datapath, b"data2")
    temp_branch.commit(message="Update data.txt")

    assert fs.exists(datapath)
    assert fs.read_text(datapath) == "data2"

    # caret at the end of the ref should point to the first descendant...
    previous_datapath = prefix + "^/data.txt"
    assert fs.read_text(previous_datapath) == "data1"

    # ...and, since we have a linear history, equal "~".
    previous_with_tilde = previous_datapath.replace("^", "~")
    assert fs.read_text(previous_with_tilde) == "data1"



"""
Functionality for extended lakeFS transactions to conduct versioning operations between file uploads.
"""

import logging
import random
import string
import warnings
from collections import deque
from typing import TYPE_CHECKING, Literal, TypeVar

import lakefs
from fsspec.transaction import Transaction
from lakefs.branch import Branch, Reference
from lakefs.client import Client
from lakefs.exceptions import ServerException
from lakefs.object import ObjectWriter
from lakefs.reference import Commit, ReferenceType
from lakefs.repository import Repository
from lakefs.tag import Tag

T = TypeVar("T")

logger = logging.getLogger("lakefs-spec")

if TYPE_CHECKING:  # pragma: no cover
    from lakefs_spec import LakeFSFileSystem


def _ensurebranch(b: str | Branch, repository: str, client: Client) -> Branch:
    if isinstance(b, str):
        return Branch(repository, b, client=client)
    return b


class LakeFSTransaction(Transaction):
    """
    A lakeFS transaction model capable of versioning operations in between file uploads.

    Parameters
    ----------
    fs: LakeFSFileSystem
        The lakeFS file system associated with the transaction.
    """

    def __init__(self, fs: "LakeFSFileSystem"):
        super().__init__(fs=fs)
        self.fs: LakeFSFileSystem
        self.files: deque[ObjectWriter] = deque(self.files)

        self.repository: str | None = None
        self.base_branch: Branch | None = None
        self.automerge: bool = False
        self.delete: Literal["onsuccess", "always", "never"] = "onsuccess"
        self.squash: bool = False
        self._ephemeral_branch: Branch | None = None

    def __call__(
        self,
        repository: str | Repository,
        base_branch: str | Branch = "main",
        branch_name: str | None = None,
        automerge: bool = True,
        delete: Literal["onsuccess", "always", "never"] = "onsuccess",
        squash: bool = False,
    ) -> "LakeFSTransaction":
        """
        Creates an ephemeral branch, conducts all uploads and operations on that branch,
        and optionally merges it back into the source branch.

        repository: str | Repository
            The repository in which to conduct the transaction.
        base_branch: str | Branch
            The branch on which the transaction operations should be based.
        automerge: bool
            Automatically merge the ephemeral branch into the base branch after successful
            transaction completion.
        delete: Literal["onsuccess", "always", "never"]
            Cleanup policy / deletion handling for the ephemeral branch after the transaction.

            If ``"onsuccess"``, the branch is deleted if the transaction succeeded,
            or left over if an error occurred.

            If ``"always"``, the ephemeral branch is always deleted after transaction regardless of success
            or failure.

            If ``"never"``, the transaction branch is always left in the repository.
        squash: bool
            Optionally squash-merges the transaction branch into the base branch.
        """

        if isinstance(repository, str):
            self.repository = repository
        else:
            self.repository = repository.id

        repo = lakefs.Repository(self.repository, client=self.fs.client)
        try:
            _ = repo.metadata
        except ServerException:
            raise ValueError(f"repository {self.repository!r} does not exist") from None

        # base branch needs to be a lakefs.Branch, since it is being diffed
        # with the ephemeral branch in __exit__.
        self.base_branch = _ensurebranch(base_branch, self.repository, self.fs.client)

        self.automerge = automerge
        self.delete = delete
        self.squash = squash

        ephem_name = branch_name or "transaction-" + "".join(random.choices(string.digits, k=6))  # noqa: S311
        self._ephemeral_branch = Branch(self.repository, ephem_name, client=self.fs.client)
        return self

    def __enter__(self):
        logger.debug(
            f"Creating ephemeral branch {self._ephemeral_branch.id!r} "
            f"from branch {self.base_branch.id!r}."
        )
        self._ephemeral_branch.create(self.base_branch, exist_ok=False)
        self.fs._intrans = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        success = exc_type is None
        while self.files:
            # fsspec base class calls `append` on the file, which means we
            # have to pop from the left to preserve order.
            f = self.files.popleft()
            if not success:
                f.discard()

        self.fs._intrans = False
        self.fs._transaction = None

        if any(self._ephemeral_branch.uncommitted()):
            msg = f"Finished transaction on branch {self._ephemeral_branch.id!r} with uncommitted changes."
            if self.delete != "never":
                msg += " Objects added but not committed are lost."
            warnings.warn(msg)

        if success and self.automerge:
            if any(self.base_branch.diff(self._ephemeral_branch)):
                self._ephemeral_branch.merge_into(self.base_branch, squash_merge=self.squash)
        if self.delete == "always" or (success and self.delete == "onsuccess"):
            self._ephemeral_branch.delete()

    @property
    def branch(self):
        return self._ephemeral_branch

    def commit(self, message: str, metadata: dict[str, str] | None = None) -> Reference:
        """
        Create a commit on this transaction's ephemeral branch with a commit message
        and attached metadata.

        Parameters
        ----------
        message: str
            The commit message to attach to the newly created commit.
        metadata: dict[str, str] | None
            Optional metadata to enrich the created commit with (author, e-mail, ...).

        Returns
        -------
        Reference
            The created commit.
        """

        diff = list(self.branch.uncommitted())

        if not diff:
            logger.warning(f"No changes to commit on branch {self.branch.id!r}.")
            return self.branch.head

        return self.branch.commit(message, metadata=metadata)

    def merge(self, source_ref: str | Branch, into: str | Branch, squash: bool = False) -> Commit:
        """
        Merge a branch into another branch in a repository.

        In case the branch contains no changes relevant to the target branch,
        no merge happens, and the tip of the target branch is returned instead.

        Parameters
        ----------
        source_ref: str | Branch
            Source reference containing the changes to merge.
            Can be a branch name or partial commit SHA.
        into: str | Branch
            Target branch into which the changes will be merged.
        squash: bool
            Optionally squash-merges the source reference into the target branch.

        Returns
        -------
        Commit
            Either the created merge commit, or the head commit of the target branch.
        """
        source = _ensurebranch(source_ref, self.repository, self.fs.client)
        dest = _ensurebranch(into, self.repository, self.fs.client)

        if any(dest.diff(source)):
            source.merge_into(dest, squash_merge=squash)
        return dest.head.get_commit()

    def revert(self, branch: str | Branch, ref: ReferenceType, parent_number: int = 1) -> Commit:
        """
        Revert a previous commit on a branch.

        Parameters
        ----------
        branch: str | Branch
            Branch on which the commit should be reverted.
        ref: ReferenceType
            The reference to revert.
        parent_number: int
            If there are multiple parents to a commit, specify to which parent
            the commit should be reverted. ``parent_number = 1`` (the default)
            refers to the first parent commit of the current ``branch`` tip.

        Returns
        -------
        Commit
            The created revert commit.
        """

        b = _ensurebranch(branch, self.repository, self.fs.client)

        ref_id = ref if isinstance(ref, str) else ref.id
        b.revert(ref_id, parent_number=parent_number)
        return b.head.get_commit()

    def rev_parse(self, ref: ReferenceType) -> Commit:
        """
        Parse a given lakeFS reference expression and obtain its corresponding commit.

        Parameters
        ----------
        ref: ReferenceType
            Reference object to resolve, can be a branch, commit SHA, or tag.

        Returns
        -------
        Commit
            The commit referenced by the expression ``ref``.
        """

        ref_id = ref.id if isinstance(ref, Reference) else ref
        reference = lakefs.Reference(self.repository, ref_id, client=self.fs.client)
        return reference.get_commit()

    def tag(self, ref: ReferenceType, name: str) -> Tag:
        """
        Create a tag referencing a commit in a repository.

        Parameters
        ----------
        ref: ReferenceType
            Commit SHA or placeholder for a reference or commit object
            to which the new tag will point.
        name: str
            Name of the tag to be created.

        Returns
        -------
        Tag
            The requested tag.
        """

        return lakefs.Tag(self.repository, name, client=self.fs.client).create(ref)



# Functional syntax to allow for the attribute name containing a dash
from typing import Literal, TypedDict

from typing_extensions import Required

ObjectType = Literal["file", "directory"]

ObjectInfoData = TypedDict(
    "ObjectInfoData",
    {
        "type": Required[ObjectType],
        "name": Required[str],
        "size": int | None,
        "checksum": str | None,
        # TODO: the dash was an unfortunate choice, but is kept for backwards compatibility.
        "content-type": str | None,
        "mtime": int | None,
        "metadata": dict[str, str] | None,
    },
    total=False,
)



from lakefs import Branch, Repository

from lakefs_spec.spec import LakeFSFileSystem


def test_gh_319(
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    """
    Regression test for GitHub issue 319: lakefs_spec does not display custom metadata
    https://github.com/aai-institute/lakefs-spec/issues/319
    """

    rpath = f"lakefs://{repository.id}/{temp_branch.id}/test.txt"

    # Create a file with custom metadata
    fs.pipe(
        rpath,
        b"Hello, world!",
        metadata={"custom_key": "custom_value"},
    )

    # Read the file and check the custom metadata
    info = fs.info(rpath)

    metadata = info.get("metadata")
    assert metadata is not None, "Metadata should not be None"
    assert metadata.get("custom_key") == "custom_value"



from lakefs import Branch, Repository

from lakefs_spec.spec import LakeFSFileSystem


def test_gh_314(
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    """
    Regression test for GitHub issue 314: Enable `@` and `~N` syntax
    https://github.com/aai-institute/lakefs-spec/issues/314
    """

    prefix = f"lakefs://{repository.id}/{temp_branch.id}"
    datapath = f"{prefix}/data.txt"

    # add new file, and immediately commit.
    fs.pipe(datapath, b"data1")
    temp_branch.commit(message="Add data.txt")

    fs.pipe(datapath, b"data2")
    # Reading the committed version of the file should yield the correct data.
    committed_head_path = f"{prefix}@/data.txt"
    assert fs.read_text(committed_head_path) == "data1"

    # Reading a relative commit should yield the correct data.
    temp_branch.commit(message="Update data.txt")
    relative_commit_path = f"{prefix}~1/data.txt"
    assert fs.read_text(relative_commit_path) == "data1"



import re

import pytest

from lakefs_spec.util import _batched, _uri_parts


def test_batched_empty_iterable():
    result = list(_batched([], 3))
    assert result == []


def test_batched_single_batch():
    result = list(_batched([1, 2, 3], 3))
    assert result == [(1, 2, 3)]


def test_batched_multiple_batches():
    result = list(_batched([1, 2, 3, 4, 5, 6, 7], 3))
    assert result == [(1, 2, 3), (4, 5, 6), (7,)]


def test_batched_batch_size_greater_than_iterable():
    result = list(_batched([1, 2], 5))
    assert result == [(1, 2)]


def test_batched_invalid_batch_size():
    with pytest.raises(ValueError, match="n must be at least one"):
        list(_batched([1, 2, 3], 0))


class TestLakeFSUriPartRegexes:
    @pytest.mark.parametrize(
        "repo_name, valid",
        [
            ("my-repo", True),
            ("@@repo", False),
            ("", False),
            ("a", False),
            ("a" * 63, True),
            ("a" * 64, False),
        ],
    )
    def test_repository(self, repo_name: str, valid: bool) -> None:
        result = re.match(_uri_parts["repository"], repo_name + "/")
        if valid:
            assert result is not None
        else:
            assert result is None

    @pytest.mark.parametrize(
        "refexp, valid",
        [
            ("", False),
            ("main", True),
            ("main@", True),
            ("main~", True),
            ("main^", True),
            ("main^2", True),
            ("main^^^", True),
            ("main^1^1", True),
            ("main^1~1", True),
        ],
    )
    def test_ref_expression(self, refexp: str, valid: bool) -> None:
        result = re.match(_uri_parts["ref expression"], refexp + "/")
        if valid:
            assert result is not None
        else:
            assert result is None



import os
from pathlib import Path
from typing import TypeAlias

import pytest

from lakefs_spec import LakeFSFileSystem

AnyPath: TypeAlias = str | os.PathLike[str] | Path


@pytest.mark.parametrize(
    "path,expected",
    [
        # base case
        ("lakefs://repo/ref/a.txt", "repo/ref/a.txt"),
        # does not strip trailing slash
        ("lakefs://repo/ref/", "repo/ref/"),
        # can accept Path
        (Path("repo/ref/a.txt"), "repo/ref/a.txt"),
        # Works on lists as well
        (
            ["lakefs://repo/ref/a.txt", "lakefs://repo/ref/b.txt"],
            ["repo/ref/a.txt", "repo/ref/b.txt"],
        ),
    ],
)
def test_strip_protocol(
    fs: LakeFSFileSystem,
    path: AnyPath | list[AnyPath],
    expected: str | list[str],
) -> None:
    actual = fs._strip_protocol(path)
    assert actual == expected



from lakefs import Branch, Repository

from lakefs_spec.spec import LakeFSFileSystem


def test_gh_321(
    fs: LakeFSFileSystem,
    repository: Repository,
    temp_branch: Branch,
) -> None:
    """
    Regression test for GitHub issue 321: ls() method doesn't return the expected type in info
    https://github.com/aai-institute/lakefs-spec/issues/321
    """

    from pyarrow.fs import FileSelector, FileType, FSSpecHandler, PyFileSystem

    rpath = f"lakefs://{repository.id}/{temp_branch.id}"

    # Use the fs object to list files in the directory
    pa_lakefs = PyFileSystem(FSSpecHandler(fs))
    info = pa_lakefs.get_file_info(FileSelector(f"{rpath}/data", recursive=True))

    for item in info:
        assert item.type != FileType.Unknown
